package org.n52.wmsclientcore;

import java.awt.image.renderable.ParameterBlock;
import java.io.Serializable;
import java.net.URL;
import java.util.ArrayList;
import java.util.ListIterator;

import javax.media.jai.JAI;
import javax.media.jai.PlanarImage;
import javax.media.jai.RenderedOp;

import org.apache.log4j.Logger;

import de.ifgi.shared.jai.CMConvertDescriptor;
import de.ifgi.shared.jai.MosaikDescriptor;
import de.ifgi.shared.jai.TransparencyDescriptor;



/**
 * @author @author <a href="mailto:senkler@ifgi.uni-muenster.de">Kristian Senkler</a>
 *
 * Class to handle image overlay. By the time, the image format is restricted to
 * PNG images. The images are either accessible via a URL or can be handed to
 * the class as RendereOp Instances.
 *
 * The order of the added images is FIFO.
 *
 * TBD:
 * - Exception handling, if an image could not be read
 */
public class ImageOverlay implements Serializable {

    private static Logger logger = Logger.getLogger(ImageOverlay.class.getName());

    private ArrayList imageList;

    /**
     * Constructor for ImageOverlay.
     */
    public ImageOverlay() {
        //register descriptors
        CMConvertDescriptor.register();
        TransparencyDescriptor.register();
        MosaikDescriptor.register();

        this.imageList = new ArrayList();
    }

    /**
     * Method addImage. Adds a PlanarImage to the List
     * @param inImage
     */
    public void addImage(PlanarImage inImage) {
    	    this.imageList.add(inImage);
    }

    /**
     * Method addImageURL. Adds an URL (i.e. GetMap-Request) to the List
     * Method processes automatically the image
     */
    public void addImageURL(String inURL) throws Exception {
        URL requestURL = new URL(inURL);
        RenderedOp img = JAI.create("url", requestURL);
        this.imageList.add(img.createInstance());
    }

    /**
     * Method clearImageList. Clears the imageList.
     */
    public void clearImageList() {
        this.imageList.clear();
    }

    /**
     * Method doOverlay. Overlays every Image in the corresponding imageList
     * from bottom to top.
     * @return RenderedImage
     */
    public PlanarImage doOverlay() {

        //iterate through imageList
        ListIterator itr = this.imageList.listIterator();
        ParameterBlock entireBlock = new ParameterBlock();
        PlanarImage currentImage;
        ParameterBlock currentBlock;

        logger.debug("Starting map overlay ...");
        while (itr.hasNext()) {
            // handle transparency (default=white)
            currentBlock = new ParameterBlock();
            currentImage = (PlanarImage) itr.next();
            currentBlock.addSource(currentImage);
            currentImage = JAI.create("ifgiTransparency", currentBlock);
/*
            // set color models to RGB by default
            if (currentImage.getColorModel() != null) {
                if (!currentImage.getColorModel().equals(ColorModel.getRGBdefault())) {
                	IndexColorModel icm = (IndexColorModel) currentImage.getColorModel();
                	StringBuffer sb = new StringBuffer(10000);
                	for (int i = 0; i < icm.getMapSize(); i++) {
						int rgb = icm.getRGB(i);
						sb.append(Integer.toHexString(rgb));
						sb.append(" | ");
                	}
                	LOGGER.debug("Handling server image: ICM " + sb.toString());
                    ParameterBlock pb1 = new ParameterBlock();
                    pb1.addSource(currentImage);
                    pb1.add(ColorModel.getRGBdefault());
                    currentImage = JAI.create("ifgiCMConvert", pb1);
                }
            }
            // overlay images
             */
            entireBlock.addSource(currentImage);
        }
        currentImage = JAI.create("ifgiMosaik", entireBlock);
        logger.debug("Map overlay constructed");
        return currentImage;
    }
}
