/*
 * Created on 31.03.2003
 *
 * To change this generated comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package org.n52.wmsclientcore;
import java.awt.Point;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.io.Serializable;

import org.n52.wmsclientcore.capabilities.model.BoundingBox;
import org.n52.wmsclientcore.capabilities.model.EX_GeographicBoundingBox;


/**
 * Stellt einen Kartenausschnitt in Form eines Rechteckes dar, das durch den linken, untern Punkt und den rechten, oberen Punkt
 * eindeutig beschrieben ist.
 * @version 0.3, 23 Mar 2003
 * @author <a href="mailto:stemeyer@ifgi.uni-muenster.de">Sascha Tegtmeyer</a>
 * @author <a href="mailto:sahlmann@ifgi.uni-muenster.de">Arnd Sahlmann</a>
 * @author <a href="mailto:drewnak@ifgi.uni-muenster.de">Jan Drewnak</a>
 */
public class WmcBoundingBox implements Cloneable, Serializable {
	/**
	 * Angefragter Ausschnitt wird an die kleinere Ausdehung angepasst; das fhrt insbesondere beim initialen Request dazu, dass die angefragte BoundingBox immer innerhalb der Gesamt-BoundingBox eines Layers liegt. 
	 */
	public static final int FIT_MINIMUM = 0;
	public static final int FIT_INTERMEDIATE = 1;
	public static final int FIT_MAXIMUM = 2;
	private double resX_;
	private double resY_;
	private String srs_;
	private Rectangle2D actualBBox;

	/**
	 * Es wird eine neue Boundingbox erstellt unter Angabe des linken unteren Punktes und des rechten oberen.
	 * @param lowerLeft Rechts- und Hochwert des linken unteren Punktes
	 * @param upperRight Rechts- und Hochwert des rechten oberen Punktes
	 */
	public WmcBoundingBox(Point2D lowerLeft, Point2D upperRight) {
		this("", lowerLeft, upperRight);
	}

	/**
	 * Bounding Box with corresponding SRS and Resolution in x resp. y.
	 * Note: resolution is an optional parameter according to OGC WMS 1.1.0
	 * @param inSRS Spatial Reference System zur BoundingBox
	 * @param lowerLeft Rechts- und Hochwert des linken unteren Punktes
	 * @param upperRight Rechts- und Hochwert des rechten oberen Punktes
	 */
	public WmcBoundingBox(String inSRS, Point2D lowerLeft, Point2D upperRight) {
		this(inSRS, lowerLeft, upperRight, 0.0, 0.0);
	}
	public WmcBoundingBox(BoundingBox box) {
		this(box.getCRS(), new Point2D.Double(box.getMinx(), box.getMiny()), new Point2D.Double(box.getMaxx(), box.getMaxy()));
	}
	public WmcBoundingBox(org.n52.wmsclientcore.context.BoundingBox box)
	{
		this(box.getSRS(), new Point2D.Double(box.getMinx().doubleValue(), box.getMiny().doubleValue()), new Point2D.Double(box.getMaxx().doubleValue(), box.getMaxy().doubleValue()));
	}
	/*
	 * this constructor sets the SRS static 4326, eg. WMS Spec
	 */
	public WmcBoundingBox(EX_GeographicBoundingBox box) {
		this("EPSG:4326", new Point.Double(box.getEastBoundLongitude(), box.getSouthBoundLatitude()), new Point2D.Double(box.getWestBoundLongitude(), box.getNorthBoundLatitude()));
	}
	/**
	 * Bounding Box with corresponding SRS and Resolution in x resp. y.
	 * Note: resolution is an optional parameter according to OGC WMS 1.1.0
	 * @param inSRS Spatial Reference System zur BoundingBox
	 * @param lowerLeft Rechts- und Hochwert des linken unteren Punktes
	 * @param upperRight Rechts- und Hochwert des rechten oberen Punktes
	 * @param inResX Auflsung X
	 * @param inResY Auflsung Y
	 */
	public WmcBoundingBox(String inSRS, Point2D lowerLeft, Point2D upperRight, double inResX, double inResY) {
		resX_ = inResX;
		resY_ = inResY;
		srs_ = inSRS;
		buildRectangle(lowerLeft, upperRight);
	}

	private void buildRectangle(Point2D lowerLeft, Point2D upperRight) {
		this.actualBBox = new Rectangle2D.Double(lowerLeft.getX(), lowerLeft.getY(), upperRight.getX() - lowerLeft.getX(), upperRight.getY() - lowerLeft.getY());

	}

	/**
	 * Setz neue Koordinaten fr die Boundingbox.
	 * @param lowerLeft Rechts- und Hochwert des linken unteren Punktes
	 * @param upperRight Rechts- und Hochwert des rechten oberen Punktes
	 */
	public void setBBox(Point2D lowerLeft, Point2D upperRight) {
		buildRectangle(lowerLeft, lowerLeft);
	}

	/**
	 * Gibt den unteren linken Punkt der Boundingbox zurck.
	 * @return Unteren linken Punkt der Boundingbox
	 */
	public Point2D getLowerLeft() {
		return new Point2D.Double(actualBBox.getMinX(), actualBBox.getMinY());
	}

	/**
	 * Gibt den oberen rechten Punkt der Boundingbox zurck.
	 * @return Oberen rechten Punkt der Boundingbox
	 */
	public Point2D getUpperRight() {
		return new Point2D.Double(actualBBox.getMaxX(), actualBBox.getMaxY());
	}

	/**
	 * Gibt die BoundingBox als java.awt.geom.Rectangle2D zurck.
	 * @return eine Reprsentation der BoundingBox als java.awt.geom.Rectangle2D
	 */
	public Rectangle2D getActualBBox() {
		return actualBBox;
	}

	/**
	 * Gibt eine String-Rersentation der Boundingbox zurck.
	 * @return String im Format: "minX,minY,maxX,maxY" - getrennt durch Kommata;
	 */
	public String getBBoxString() {
		return String.valueOf(actualBBox.getMinX()) + "," + String.valueOf(actualBBox.getMinY()) + "," + String.valueOf(actualBBox.getMaxX()) + "," + String.valueOf(actualBBox.getMaxY());
	}

	/**
	 * Fhrt eine Vergrerung/Verkleinerung der BoundingBox durch. 
	 * @param newCenter neuer Mittelpunkt der BoundingBox. Wenn center den Wert null besitzt, wird der aktuelle Mittelpunkt der BoundingBox verwendet  
	 * @param factor Zoomfaktor z. Fr z > 1 wird um den Faktor z in die BoundingBox hereingezommt (entspricht einer Vergrerung des Mastabs bzw. Verkleinerung der BoundingBox), fr z < 1 wird um den Factor 1/z herausgezommt (entspricht einer Verkleinerung des Mastabs bzw. Vergrerung der BoundingBox)
	 */
	public void zoom(Point2D newCenter, double factor) {
		AffineTransform at = new AffineTransform();
		double bboxCenterX = actualBBox.getCenterX();
		double bboxCenterY = actualBBox.getCenterY();
		if (newCenter == null)
			newCenter = new Point2D.Double(bboxCenterX, bboxCenterY);
		double centerX = newCenter.getX();
		double centerY = newCenter.getY();
		factor = 1 / factor;
		at.translate(centerX, centerY);
		at.scale(factor, factor);			
		at.translate(-bboxCenterX, -bboxCenterY);
		actualBBox = at.createTransformedShape(actualBBox).getBounds2D();
	}
	
	/**
		 * Fhrt eine Vergrerung der BoundingBox durch Aufziehen eines Rechtecks durch. 
		 * @param newCenter neuer Mittelpunkt der BoundingBox. Wenn center den Wert null besitzt, wird der aktuelle Mittelpunkt der BoundingBox verwendet  
		 * @param factorX Zoomfaktor angepasst an die jeweilige Zoomstufe der Kartenbreite
		 * @param factorY Zoomfaktor angepasst an die jeweilige Zoomstufe der Kartenhhe
		 */
		public void zoom(Point2D newCenter, double factorX, double factorY) {
			AffineTransform at = new AffineTransform();
			double bboxCenterX = actualBBox.getCenterX();
			double bboxCenterY = actualBBox.getCenterY();
			if (newCenter == null)
				newCenter = new Point2D.Double(bboxCenterX, bboxCenterY);
			double centerX = newCenter.getX();
			double centerY = newCenter.getY();
			factorX = 1 / factorX;
			factorY = 1 / factorY;
			at.translate(centerX, centerY);
			at.scale(factorX, factorY);
			at.translate(-bboxCenterX, -bboxCenterY);
			actualBBox = at.createTransformedShape(actualBBox).getBounds2D();
		}
	
//	/**
//		 * Fhrt eine Vergrerung der BoundingBox durch Aufziehen eines Rechtecks auf der Karte aus. 
//		 * @param newCenter Zentrum der aufgezogenen Bounding Box.   
//		 * @param width Breite des aufgezogenen Rechtecks
//		 * @param height Hhe des aufgezogenen Rechtecks
//		 */
//		public void zoom(int screenX, int screenY, int width, int height) {
//			if (width == height) {
//				boxWidth = width;
//				boxHeight = width;
//			}
//			else if (width < height && actualBBox.getWidth() < actualBBox.getHeight()) {
//				double boxScaleFactor = actualBBox.getHeight()/ actualBBox.getWidth();
//				boxWidth = width;
//				boxHeight = width * boxScaleFactor;
//			}
//			else if (width < height && actualBBox.getWidth() > actualBBox.getHeight()) {
//				double boxScaleFactor = actualBBox.getWidth()/ actualBBox.getHeight();
//				boxWidth = width;
//				boxHeight = width * boxScaleFactor;
//			}
//			else if (width > height && actualBBox.getWidth() < actualBBox.getHeight()) {
//				double boxScaleFactor = actualBBox.getWidth()/ actualBBox.getHeight();
//				boxWidth = height * boxScaleFactor;
//				boxHeight = height;
//			}
//			else if (width > height && actualBBox.getWidth() > actualBBox.getHeight()) {
//				double boxScaleFactor = actualBBox.getHeight()/ actualBBox.getWidth();
//				boxWidth = height * boxScaleFactor;
//				boxHeight = height;
//			}	
//			else if (width < height && actualBBox.getWidth() == actualBBox.getHeight()) {
//				boxWidth = width;
//				boxHeight = width;
//			}	
//			else if (width > height && actualBBox.getWidth() == actualBBox.getHeight()) {				
//				boxWidth = height;
//				boxHeight = height;
//			}
//			else if (width == height && actualBBox.getWidth() == actualBBox.getHeight()) {
//				boxWidth = width;
//				boxHeight = width;
//			}
//			
//			double centerX = screenX + width / 2;
//			double centerY = screenY + height / 2;
//			
//			double upperLeftX = centerX - boxWidth / 2;
//			double upperLeftY = centerY - boxHeight / 2;
//			Point2D upperLeft = screen2Realworld((int)actualBBox.getWidth(), (int)actualBBox.getHeight(), new java.awt.Point((int)upperLeftX, (int)upperLeftY));
//			
//			double lowerRightX = centerX + boxWidth / 2;
//			double lowerRightY = centerY + boxHeight / 2;	
//			Point2D lowerRight = screen2Realworld((int)actualBBox.getWidth(), (int)actualBBox.getHeight(), new java.awt.Point((int)lowerRightX, (int)lowerRightY));			
//			buildRectangle(upperLeft, lowerRight);															
//		}

	/**
	 * Verschiebt die BoundingBox
	 * 
	 * @param dirX Werte: -1: Verschieben in negative X-Achsen-Richtung<br>0: kein Verschieben entlang der X-Achse<br>1: Verschieben in positive X-Achsen-Richtung
	 * @param dirY Werte: -1: Verschieben in negative Y-Achsen-Richtung<br>0: kein Verschieben entlang der Y-Achse<br>1: Verschieben in positive Y-Achsen-Richtung
	 * @param factorX Faktor x, um den die BoundingBox in X-Achsen-Richtung verschoben werden soll. Bsp.: x==0.5 versetzt den Mittelpunkt an den Rand der X-Achsen-Ausdehnung  
	 * @param factorY Faktor y, um den die BoundingBox in Y-Achsen-Richtung verschoben werden soll. Bsp.: y==1 versetzt den Mittelpunkt um die aktuelle y-Achsen-Ausdehnung
	 */
	public void pan(int dirX, int dirY, double factorX, double factorY) {
		AffineTransform at = new AffineTransform();
		double offsetX = actualBBox.getWidth() * factorX * dirX;
		double offsetY = actualBBox.getHeight() * factorY * dirY;
		at.translate(offsetX, offsetY);
		actualBBox = at.createTransformedShape(actualBBox).getBounds2D();
	}

	public void pan(String nav, double factor) {
		int xDir = 0;
		int yDir = 0;

		if (nav.equals("north"))
			yDir = 1;
		else if (nav.equals("northeast")) {
			xDir = 1;
			yDir = 1;
		} else if (nav.equals("northwest")) {
			xDir = -1;
			yDir = 1;
		} else if (nav.equals("south"))
			yDir = -1;
		else if (nav.equals("southeast")) {
			xDir = 1;
			yDir = -1;
		} else if (nav.equals("southwest")) {
			xDir = -1;
			yDir = -1;
		} else if (nav.equals("west"))
			xDir = -1;
		else if (nav.equals("east"))
			xDir = 1;
		pan(xDir, yDir, factor, factor);
	}
	/**
	 * Transformiert einen Punkt von einem (Bild-)Koordinatensystem in das Koordinatensystem der BoundingBox
	 * 
	 * @param screenWidth Breite des Bildes
	 * @param screenHeight Hhe des Bildes
	 * @param screenPoint screen coordinates (origin = upper left corner)
	 */
	public Point2D screen2Realworld(int screenWidth, int screenHeight, Point screenPoint) {
		screenPoint.y = screenHeight - screenPoint.y;
		Point2D.Double tPoint = new Point2D.Double();
		AffineTransform at = new AffineTransform();
		double factorX = screenWidth / actualBBox.getWidth();
		double factorY = screenHeight / actualBBox.getHeight();
		at.scale(factorX, factorY);
		at.translate(-actualBBox.getX(), -actualBBox.getY());
		try {
			at.createInverse().transform(screenPoint, tPoint);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return tPoint;
	}

	/**
	 * Passt die BoundingBox so an, dass sie das gleiche Seitenverhltnis wie das des Ausgabebildes besitzt
	 * 
	 * @param screenWidth Breite des Ausgabebildes
	 * @param screenHeight Hhe des Ausgabebildes
	 * @param fitPolicy
	 */
	public void fitBBox2Screen(int screenWidth, int screenHeight, int fitPolicy) {
		double aspectRatio = screenWidth / screenHeight;
		double tW = screenWidth * actualBBox.getHeight(); //factorY
		double tH = screenHeight * actualBBox.getWidth(); //factorX
		double factorX = 1 / tH;
		double factorY = 1 / tW;
		AffineTransform at = new AffineTransform();
		at.translate(actualBBox.getX(), actualBBox.getY());

		switch (fitPolicy) {
			case FIT_MINIMUM :
				at.scale(Math.min(tH, tW), Math.min(tH, tW));
				break;
			case FIT_MAXIMUM :
				at.scale(Math.max(tH, tW), Math.max(tH, tW));
				break;
			case FIT_INTERMEDIATE :
				{
					if (aspectRatio > 1)
						at.scale(tH, tH);
					else
						at.scale(tW, tW);
					break;
				}
			default :
				at.scale(Math.min(tH, tW), Math.min(tH, tW));
				break;
		}

		/*
				if (aspectRatio > 1) at.scale(tH,tH);
				else at.scale(tW,tW);
		*/

		at.scale(factorX, factorY);
		at.translate(-actualBBox.getX(), -actualBBox.getY());
		actualBBox = at.createTransformedShape(actualBBox).getBounds2D();
	}

	public void union(WmcBoundingBox tBox) {
		Rectangle2D.union(actualBBox, tBox.getActualBBox(), actualBBox);
	}

	public double getXScale(double screenWidth) {
		return actualBBox.getWidth() / screenWidth;
	}
	public double getYScale(double screenHeight) {
		return actualBBox.getHeight() / screenHeight;
	}
	public double getMaxScale(int screenWidth, int screenHeight) {
		return Math.max(getXScale(screenWidth), getYScale(screenHeight));
	}
	public double getMinScale(int screenWidth, int screenHeight) {
		return Math.min(getXScale(screenWidth), getYScale(screenHeight));
	}
	/** sets spatial reference system */
	public void setSRS(String inSRS) {
		srs_ = inSRS;
	}

	/** returns spatial reference system */
	public String getSRS() {
		return srs_;
	}

	/** sets resolution in x direction */
	public void setResX(double inResX) {
		resX_ = inResX;
	}

	/** returns resolution in x direction */
	public double getResX() {
		return resX_;
	}

	/** sets resolution in y direction */
	public void setResY(double inResY) {
		resY_ = inResY;
	}

	/** returns resolution in y direction */
	public double getResY() {
		return resY_;
	}

	public String toString() {
		return getBBoxString();
	}

	public Object clone() {
		try {
			WmcBoundingBox aBBox = (WmcBoundingBox) super.clone();
			//			aBBox.lowerLeft = (Point2D.Double) this.lowerLeft.clone();
			//			aBBox.upperRight = (Point2D.Double) this.upperRight.clone();
			aBBox.actualBBox = (Rectangle2D) this.actualBBox.clone();
			aBBox.srs_ = new String(srs_);
			return (Object) aBBox;
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return null;
	}
	public double currentScaleHint(int width, int height) {
		///TODO: Scale hint richtig berechenen

		/*
						double lengthOfPixelX = (bbox.getLowerLeftAsDoublePoint().getX()
										- bbox.getUpperRightAsDoublePoint().getX())
										/ Double.parseDouble(width);
		
						double lengthOfPixelY = (bbox.getLowerLeftAsDoublePoint().getY()
										- bbox.getUpperRightAsDoublePoint().getY())
										/ Double.parseDouble(height);
		*/
		double lengthOfPixelX = this.getXScale(width);

		double lengthOfPixelY = this.getYScale(height);
		double diagPixel = Math.sqrt(Math.pow(lengthOfPixelX, 2.0) + Math.pow(lengthOfPixelY, 2.0));
		return diagPixel;
	}
}